/*
  This is the Prometheus implementation of LinearSystemCore.
  $Id: Prometheus_LinSysCore.h,v 1.23 2004/08/25 19:57:43 adams Exp $
  */

#ifndef _Prometheus_LinSysCore_h_
#define _Prometheus_LinSysCore_h_

#include <mpi.h>

#if !defined(PROM_NO_FEI)

/******************************** Sierra ********************************/
#include <fei_defs.h> 
#include <base/Data.h>
#include <base/LinSysCore_flexible.h>

#else

/******************************* Olympus ********************************/
#include <stdio.h>
class PromOptions;
class PromPerfMonitor;
class PromComm;
typedef int GlobalID;
typedef struct{
  void setTypeName( char matrix_name[] ){}
  void setDataPtr( void * p ){ }
  void *getDataPtr()const{ return NULL; }
  char *getTypeName() const {return "K";}
}Data;
typedef int Lookup;
class LinearSystemCore
{
 public:
  LinearSystemCore(){}
  LinearSystemCore( MPI_Comm ){}
  virtual int parameters( int numParams, char** params ) = 0;
  virtual int setGlobalOffsets( int len, const int* nodeOffsets, 
				const int* eqnOffsets,
				const int* blkEqnOffsets) = 0;
  virtual int setConnectivities( GlobalID elemBlock,
				 int numElements,
				 int numNodesPerElem,
				 const GlobalID* elemIDs,
				 const int* const* connNodes)  = 0;
  virtual int setStiffnessMatrices(GlobalID elemBlock,
				   int numElems,
				   const GlobalID* elemIDs,
				   const double *const *const *stiff,
				   int numEqnsPerElem,
				   const int *const * eqnIndices) = 0;
  virtual int setLoadVectors( GlobalID elemBlock,
			      int numElems,
			      const GlobalID* elemIDs,
			      const double *const * load,
			      int numEqnsPerElem,
			      const int *const * eqnIndices) = 0;
  virtual int setMultCREqns( int multCRSetID,
			     int numCRs, int numNodesPerCR,
			     int** nodeNumbers, int** eqnNumbers,
			     int* fieldIDs,
			     int* multiplierEqnNumbers) = 0;

  virtual int setMatrixStructure( int** ptColIndices, int* ptRowLengths,
				  int** blkColIndices, int* blkRowLengths,
				  int* ptRowsPerBlkRow, void *mat = NULL) = 0;
  /*int resetMatrixAndVector:
    don't destroy the structure of the matrix, but set the value 's'
    throughout the matrix and vectors.
    */
  virtual int resetMatrixAndVector(double s) = 0;
  virtual int resetMatrix(double s) = 0;
  virtual int resetRHSVector(double s) = 0;

  /*int sumIntoSystemMatrix:
    this is the primary assembly function. The coefficients 'values'
    are to be accumumlated into (added to any values already in place)
    global (0-based) equation 'row' of the matrix.
  */
  virtual int sumIntoSystemMatrix( int numPtRows, const int* ptRows,
				   int numPtCols, const int* ptCols,
				   int numBlkRows, const int* blkRows,
				   int numBlkCols, const int* blkCols,
				   const double* const* values) = 0;
  
  virtual int sumIntoSystemMatrix( int numPtRows, const int* ptRows,
				   int numPtCols, const int* ptCols,
				   const double* const* values) = 0;
  /*int sumIntoRHSVector:
    this is the rhs vector equivalent to sumIntoSystemMatrix above.
  */
  virtual int sumIntoRHSVector(int num,const double*values,const int*indices)=0;
  virtual int putIntoRHSVector(int num,const double*values,const int*indices)=0;
  /* int matrixLoadComplete:
     do any internal synchronization/communication.
  */
  virtual int matrixLoadComplete() = 0;
  
  virtual int putNodalFieldData( int fieldID, int fieldSize, int* nodeNumbers,
				 int numNodes, const double* data) = 0;
  /* functions for enforcing boundary conditions. */
  virtual int enforceEssentialBC( int* globalEqn,
				  double* alpha,
				  double* gamma, int len) = 0;
  
  virtual int enforceRemoteEssBCs( int numEqns, int* globalEqns,
				   int** colIndices, int* colIndLen,
				   double** coefs) = 0;
  
  virtual int enforceOtherBC( int* globalEqn, double* alpha,
			      double* beta, double* gamma,
			      int len) = 0;
  
  /* functions for managing multiple rhs vectors */
  virtual int setNumRHSVectors(int numRHSs, const int* rhsIDs) = 0;
  
  /* virtual int setRHSID:
     set the 'current' rhs context, assuming multiple rhs vectors. */
  virtual int setRHSID(int rhsID) = 0;
  
  /*int putInitialGuess:
    function for setting (a subset of) the initial-guess
    solution values (i.e., in the 'x' vector).
  */
  virtual int putInitialGuess(const int* eqnNumbers, const double* values, int len) = 0;
  
  /* function for getting all of the answers ('x' vector). */
  virtual int getSolution(double* answers, int len) = 0;
  
  /* function for getting the (single) entry at equation
     number 'eqnNumber'. */
  virtual int getSolnEntry(int eqnNumber, double& answer) = 0;
 
  /* function for launching the linear solver */
  virtual int launchSolver( int &solveStatus, int &iterations) = 0;
};
class LinSysCore_flexible : public LinearSystemCore
{  
  // LSC_flexible
  int resetConstraints( double s ); /* LSC_flexible */
  int constraintsLoadComplete(){ return -1; }  /* (old) LSC_flexible */
  int setMultCRComplete();  /* LSC_flexible */
};  
#endif

#define PROM_GET_VALUES -3
#define PEN_INIT 1.e100

class PromCRVector;
class PromCRMatrix;
class Prometheus_EssBCData;
class Prometheus_OthBCData;
class Prometheus_ZeroEquations;
class Prometheus_Constr;
class Prometheus;
class PromVector;
class PromMatrix;
class PromGrid;
class PromMap;
template <class T> class PromList;
template <class T> class PromTable;

/*************************************************************************
 * class PromCRCache 
 */
class PromCRCache
{
 public:
  PromCRCache( int len, int geq, int* gids, double *vals );
  ~PromCRCache();
  int len_;
  int geq_;
  int *geqs_;
  double *vals_; 
  PromCRCache *next_;
};

/*************************************************************************
 * class Prometheus_EssBCData 
 */
class Prometheus_EssBCData
{
 public:
  Prometheus_EssBCData(int* globalEqn, double* alpha, double* gamma, int len);
  ~Prometheus_EssBCData();
  int len_;
  int* globalEqn_;
  double* alpha_; 
  double* gamma_;
  Prometheus_EssBCData *next_;
};

/*************************************************************************
 * class Prometheus_OthBCData 
 */
class Prometheus_OthBCData
{
 public:
  Prometheus_OthBCData( int* globalEqn, double* alpha, double* beta, 
			double* gamma, int len);
  ~Prometheus_OthBCData();
  int len_;
  int* globalEqn_;
  double* alpha_;  
  double* beta_; 
  double* gamma_;
  Prometheus_OthBCData *next_;
};

/*************************************************************************
 * class Prometheus_ZeroEquations 
 */
class Prometheus_ZeroEquations
{
public:
  Prometheus_ZeroEquations(int* localrows, int* sizes, int** idsarr, int len) :
    len_(len), sizes_(sizes), zeroRowsLeq_(localrows), idsarr_(idsarr),
    next_(NULL) {}
  ~Prometheus_ZeroEquations(){}
  int len_;
  int *sizes_;
  int *zeroRowsLeq_; 
  int **idsarr_;
  Prometheus_ZeroEquations *next_;
};

class PromCR;
class PromSolver;
//! class PromUzawaSolver
/*!
  class PromUzawaSolver:
*/
class PromUzawaSolver
{
 public:
  PromUzawaSolver() : pwork_(NULL), uwork_(NULL), uzawaC_(NULL),
    diag_(NULL), Pen_stiff_(NULL), penalty_factor_(5.0), cr_tol_(0.0),
    max_uzawa_its_(10) { }
  ~PromUzawaSolver();
  //
  int Init( const PromMap &primmap, const PromMap &dualmap );
  int Reset( const PromMap &dualmap );
  int SetReg( PromMatrix *AA );
  static int setRegularization_private( PromMatrix *const A, 
					const PromMatrix *const C,
					PromVector *const diag, 
					const double alpha,
					PromVector *const dd = NULL );
  int SolveUzawa( PromSolver *const solver, const PromMatrix *const AA,
		  PromCRVector *const RR, PromCRVector *const XX, 
		  const PromCRVector *const BB, PromVector *const x_0, 
		  const int verb, float normR_0, int &ret_out, 
		  int &its_out ) const;
  int setDummyReg( PromMatrix *const AA, 
		   const PromTable<PromCR*> *const CR_id ) const;
  // data
  PromVector    *pwork_;
  PromVector    *uwork_;
  PromMatrix    *uzawaC_;
  PromVector    *diag_;
  PromVector    *Pen_stiff_;
  double         penalty_factor_;
  double         cr_tol_;
  short int      max_uzawa_its_;
};

template <class T> class PromAList;
class PromPC;
class PromComm;
class PromOptions;
class PromPerfMonitor;
//! class Prometheus_LinSysCore: 
/*!
  class Prometheus_LinSysCore: top linear solver object
*/
class Prometheus_LinSysCore: public LinSysCore_flexible, public PromUzawaSolver
{
 public:
  Prometheus_LinSysCore( MPI_Comm comm );
  int Init( const PromComm &comm, PromOptions &opts, 
	    PromPerfMonitor &perf );
  virtual ~Prometheus_LinSysCore();
  /*for creating another one, without knowing the run-time type
    of 'this' one.    */    
  LinearSystemCore* clone();
  
  /* int parameters:
     for setting generic argc/argv style parameters.
     */
  int parameters( int numParams, char** params );
  
  int setLookup( Lookup& lookup );
  
  int setGlobalOffsets( int len, const int* nodeOffsets, const int* eqnOffsets,
			const int* blkEqnOffsets);

  int setConnectivities( GlobalID elemBlock,
			 int numElements,
			 int numNodesPerElem,
			 const GlobalID* elemIDs,
			 const int* const* connNodes) ;
  
  int setStiffnessMatrices(GlobalID elemBlock,
			    int numElems,
			    const GlobalID* elemIDs,
			    const double *const *const *stiff,
			    int numEqnsPerElem,
			    const int *const * eqnIndices);
  
  int setLoadVectors( GlobalID elemBlock,
		      int numElems,
		      const GlobalID* elemIDs,
		      const double *const * load,
		      int numEqnsPerElem,
		      const int *const * eqnIndices);
  
  int setMultCREqns( int multCRSetID,
		     int numCRs, int numNodesPerCR,
		     int** nodeNumbers, int** eqnNumbers,
		     int* fieldIDs,
		     int* multiplierEqnNumbers);
  
  int setPenCREqns( int penCRSetID,
		    int numCRs, int numNodesPerCR,
		    int** nodeNumbers, int** eqnNumbers,
		    int* fieldIDs);

  int setMatrixStructure( int** ptColIndices, int* ptRowLengths,
			  int** blkColIndices, int* blkRowLengths,
			  int* ptRowsPerBlkRow, void *mat = NULL ) ;

  /*int resetMatrixAndVector:
    don't destroy the structure of the matrix, but set the value 's'
    throughout the matrix and vectors.
    */
  int resetMatrixAndVector(double s);
  int resetMatrix(double s);
  int resetRHSVector(double s);

  /*int sumIntoSystemMatrix:
    this is the primary assembly function. The coefficients 'values'
    are to be accumumlated into (added to any values already in place)
    global (0-based) equation 'row' of the matrix.
    */
  int sumIntoSystemMatrix( int numPtRows, const int* ptRows,
			   int numPtCols, const int* ptCols,
			   int numBlkRows, const int* blkRows,
			   int numBlkCols, const int* blkCols,
			   const double* const* values);
  
  int sumIntoSystemMatrix( int numPtRows, const int* ptRows,
			   int numPtCols, const int* ptCols,
			   const double* const* values);
  
  int putIntoSystemMatrix( int numPtRows, const int* ptRows,
			   int numPtCols, const int* ptCols,
			   const double* const* values);

  int sumIntoSystemMatrix_private(int numPtRows, const int* ptRows,
				  int numPtCols, const int* ptCols,
				  const double* const* values,
				  const int add_type, const bool image);
  int getMatrixRowLength( int row, int& length);

  int getMatrixRow( int row, double* coefs, int* indices,
		    int len, int& rowLength);

  /*int sumIntoRHSVector:
    this is the rhs vector equivalent to sumIntoSystemMatrix above.
    */
  int sumIntoRHSVector(int num, const double* values, const int* indices);
  int putIntoRHSVector(int num, const double* values, const int* indices);
  int sumIntoRHSVector_private( const int num, const double* values, 
				const int* indices, const int add_type,
				const bool image );
  int accessRHSVector_private( const int num, double *values,
			       int *indices, const int add_type,
			       const bool image );

  int getFromRHSVector( int num, double* values, const int* indices);
  /* int matrixLoadComplete:
     do any internal synchronization/communication.
     */
  int matrixLoadComplete();

  int putNodalFieldData( int fieldID, int fieldSize, int* nodeNumbers,
			 int numNodes, const double* data);
 private:
  int putCoords_private( int fieldID, int fieldSize, int* nodeNumbers,
			 int numNodes, const double* data,
			 PromVector *right = NULL );
 public:
  /* functions for enforcing boundary conditions. */
  int enforceEssentialBC( int* globalEqn,
			  double* alpha,
			  double* gamma, int len);
  
  int enforceRemoteEssBCs( int numEqns, int* globalEqns,
			   int** colIndices, int* colIndLen,
			   double** coefs);
  
  int enforceOtherBC( int* globalEqn, double* alpha,
		      double* beta, double* gamma,
		      int len);
  
  /* functions for getting/setting matrix or vector pointers. */
  
  /*getMatrixPtr:
    obtain a pointer to the 'A' matrix. This should be considered a
    constant pointer -- i.e., this class remains responsible for the
    matrix (e.g., de-allocation upon destruction).
  */ 
  int getMatrixPtr(Data& data);
  
  /* copyInMatrix:
     replaces the internal matrix with a copy of the input argument, scaled
     by the coefficient 'scalar'.
  */
  int copyInMatrix(double scalar, const Data& data);
  
  /* copyOutMatrix:
     passes out a copy of the internal matrix, scaled by the coefficient
     'scalar'.
  */ 
  int copyOutMatrix(double scalar, Data& data);
  
  /* sumInMatrix:
     accumulate (sum) a copy of the input argument into the internal
     matrix, scaling the input by the coefficient 'scalar'.
  */
  int sumInMatrix(double scalar, const Data& data);
  
  /* get/setRHSVectorPtr:
     the same semantics apply here as for the matrixPtr functions above.
  */
  int getRHSVectorPtr(Data& data);
  
  /* copyInRHSVector/copyOutRHSVector/sumInRHSVector:
     the same semantics apply here as for the matrix functions above.
  */
  int copyInRHSVector(double scalar, const Data& data);
  int copyOutRHSVector(double scalar, Data& data);
  int sumInRHSVector(double scalar, const Data& data);
  
  /* destroyMatrixData/destroyVectorData:
     Utility function for destroying the matrix (or vector) in Data
  */
  int destroyMatrixData(Data& data);
  int destroyVectorData(Data& data);
  
  /* functions for managing multiple rhs vectors */
  int setNumRHSVectors(int numRHSs, const int* rhsIDs);
  
  /* int setRHSID:
     set the 'current' rhs context, assuming multiple rhs vectors. */
  int setRHSID(int rhsID);
  
  /*int putInitialGuess:
    function for setting (a subset of) the initial-guess
    solution values (i.e., in the 'x' vector).
  */
  int putInitialGuess(const int* eqnNumbers, const double* values, int len);
  
  /* function for getting all of the answers ('x' vector). */
  int getSolution(double* answers, int len);
  
  /* function for getting the (single) entry at equation
     number 'eqnNumber'. */
  int getSolnEntry(int eqnNumber, double& answer);
  
  /* function for obtaining the residual vector (using current solution or
     initial guess) */
  int formResidual(double* values, int len);

  /* function for launching the linear solver */
  int launchSolver(int& solveStatus, int& iterations);
  
  int writeSystem(const char* name);

  // LSC_flexible
  int resetConstraints( double s ); /* LSC_flexible */
  int constraintsLoadComplete(){ return -1; }  /* (old) LSC_flexible */
  int setMultCRComplete();  /* LSC_flexible */

  /* all-at-once MG support */
  int CreateKKTObs( PromGrid *const grid );
 private:
  int AggregateCRs( PromList<PromAList<int>*> &aggs, 
		    const PromMatrix*const Ti, const double tol2 = 0.2, 
		    const bool do_glob = (1==1), const bool set_with_local = (1==0) ) const;
  int MakeCRProlMap( PromList<PromAList<int>*> &aggs, 
		     const PromGrid * const grid, 
		     PromMap **clm_out, PromMatrix **prol_D_out) const; 
  int MoveAggs(const PromMap &map, 
	       PromList<PromAList<int>*> &aggs,
	       const int nextFactor, const PromMap*const CP_colMap, 
	       const PromMatrix*const CP )const;
  int SetBCforCRTable(const PromMatrix*const CC, const PromMatrix*const projC);
  int GetBCSends();
  static int Mult( const PromMatrix * const AA, const PromMatrix *const BB,
		   PromMatrix **C_out );
  static int TransMult(const PromMatrix*const AA, const PromMatrix*const BB,
		       const PromMap *const mp, PromMatrix**, 
		       const PromMap *const mpt=NULL, PromMatrix**t=NULL);
  int AggregateCRs2( PromList<PromAList<int>*> &aggs,
		     const PromMatrix *const Ti, const double tol1, 
		     const double tol2, const bool do_glob, 
		     const bool set_with_local = (1==0) ) const;
  int MIS_CRs(int &ndone, const int nadj[],
	      PromList<PromAList<int>*> &aggs, const PromMatrix *const Ti,
	      const double tol2, const bool do_glob, 
	      PromAList<int> **lid_aggList,
	      PromTable<PromAList<int>*> &ghost_list,
	      PromTable<PromAList<int>*> &listID_list,
	      int &locAggID, PromAList<int> *const dummy) const;
  int GlobalizeAggs(const PromMap &map, PromList<PromAList<int>*> &aggs,
		    PromAList<int> **lid_aggList,
		    PromTable<PromAList<int>*> &ghost_list,
		    PromTable<PromAList<int>*> &listID_list,
		    PromAList<int> *const dummy) const;
 public:
  int RemoveCRfromAllGrids();
  PromVector *getRHS();
  PromVector *getSolution();
  int SetRTol( float tol );
  int setMatrixStructure( const PromTable<int> &gid_ghostLid,
			  const int ndf ); // Olympus interface
 private:
  int setMatrixStructure_private( const PromTable<int> &gid_ghostLid,
				  const int ndf, void *mat = NULL );
  /* mapping for Langrange Multiplier problems */
  inline int eq_map_private(const int geq,int *myeq,bool *prim,int *proc)const;
  inline int id_map_private(const int gid,int *myid,bool *prim,int *proc)const;
  int primGEq_lid_private( const int geq, int *lid_out ) const;
  int id_proc_private( const int gid, int &proc) const;
  int collectBCs();
  int assembleCRs( PromTable<int> *projCREqs );
  int applyBCs( PromMatrix *projC );
  int enforceEssentialBC_private( int* globalEqn, double* alpha, double* gamma,
				  int len, PromVector *bmod );
  int enforceOtherBC_private( int* globalEqn, double* alpha,
			      double* beta, double* gamma, int len);
  int constructPrometheus( const PromTable<int> &gid_ghostLid,
			   const int ndf );
  int ExtendElemList();
  int addextraGhostsParents( PromGrid *lastg, PromGrid *nextg );
  int ReFactorSolver();
  int SymbolicSetUp();
  int MakeSolver( const PromMap *const p_map, bool ispetscpc );
  int FactorKKT( PromCRMatrix *KK, PromGrid *grid,
		 PromPC *pc, const PromCRVector *const xx);
  int LSC_Residual( PromMatrix *KK, const PromVector *const BB, 
		    PromVector *const XX, PromVector *const YY );
 public:
  int SetGlobalResTol( double tol ){ normRTol_ = tol; return 0; }
 private:
  /* data */
  PromMatrix                 *C_;       // cache for KKTMG 'C' or projecton 'C' 
  int                        *Ctnnz_;   // cache for KKTMG 'Ct' 
  PromMap                    *lmap_;    // cache for uzawa and KKTMG constraints
  PromMap                    *proj_lmap_;// cache for projection constraints
  PromVector                 **bVecArr_;
  PromVector                 *xVec_;
  PromVector                 *work_;
  PromVector                 *init_guess_;
  PromVector                 *gamma_alpha_;
  void                       *zeroIS_;
  int                        *rhsIDs_;
  int                        *proc_gnode_;
  int                        *proc_globalEq_;
  int                        *proc_lmEq_;
  int                        *proc_primEq_;
  int                        *lid_primLEq_;
  PromList<int*>             *elemList_;
  PromTable<double>          *geq_fixity_;
  PromTable<PromTable<int>*> *proc_ghostCRIDs_;
  PromTable<PromCR*>         *CR_id_;
  PromTable<PromCR*>         *projCR_id_;
  
  Prometheus_EssBCData       *essbcdata_;
  Prometheus_OthBCData       *othbcdata_;
  Prometheus_ZeroEquations   *zeroEqs_;
  PromCRCache                *CRCache_;
  PromCRCache                *projCRCache_;
 public:
  Prometheus                 *prom_;
  PromComm *Comm_; 
  PromOptions *options_;
  PromPerfMonitor *perf_mon_;
 private:
  float  normRTol_; 
  char   root_[128];
  int numGlobalRows_;
  int numGlobalRowsLm_;
  int numGlobalRowsPrim_;  
  short int currentRHS_;
  short int dirty_guess_;  /* keep track of initial solution guess in "work_"*/
 public:
  short int mg_levels_;
 private:
  short int numRHSs_;
  short int verbose_;
  short int mype_;
  short int npe_;
  short int num_matrices_;
  short int projNdLimit_;
 public:
  short int nSolves_;
  bool use_uzawa_;
  bool matrix_setup_reuse_factor_;
  bool rhs_res_scale_;

  bool use_complex_;
  static const char version_no_[32];
  MPI_Comm          MPIComm_;
  // complex methods
  int setUseComplex( bool b = true );
  bool isComplex()const{ return use_complex_; }
  int sumIntoSystemMatrix_i( int numPtRows, const int* ptRows,
			     int numPtCols, const int* ptCols,
			     int numBlkRows, const int* blkRows,
			     int numBlkCols, const int* blkCols,
			     const double* const* values);
  int sumIntoSystemMatrix_i( int numPtRows, const int* ptRows,
			     int numPtCols, const int* ptCols,
			     const double* const* values);
  int sumIntoRHSVector_i( int num, const double* values, const int* indices);
  int getFromRHSVector_i( int num, double* values, const int* indices);
  int putIntoRHSVector_i(int num, const double* values, const int* indices);
  int getSolution_i( double* answers, int len );
  int getSolnEntry_i( int eqnNumber, double& answer );
  int getSolution_private( double* answers, int len, const bool image );
  int sumIntoSystemMatrix_private(int numPtRows, const int* ptRows, 
				  int numPtCols, const int* ptCols,
				  int numBlkRows, const int* blkRows,
				  int numBlkCols, const int* blkCols,
				  const double* const* values,
				  const bool image ); 
  int getResidual_i(double* values, int len);
  int formResidual_private(double* values, int len, const bool image);
};

#endif
